// RingWorm v 1.2 (c) 1998 Graham Mitchell <mitchell@onr.com> - 19 Oct 1998

// Worm class

/*
	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License (LICENSE.TXT) for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software Foundation,
	Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "tables.h"
#include "const.h"
#include "worm.h"
#include <stdio.h>
#include <allegro.h>

enum Face { LEFT, RIGHT, TOP, BOTTOM };

Worm::Worm()
{
	// colors
	cindex = 0;
	trail = BLANK;

	// size
	len = INITIAL_LENGTH;

	// physics
	change = 9;
	friction = 0.0;

	points = 0;

	// and suitable junk initial values for those we'll overwrite in init()
	color = cbase = BLANK;
	r = angle = ident = 0;
	speed = dx = dy = 0;
	loc.x = loc.y = old_loc.x = old_loc.y = 0;
	K_LEFT = K_RIGHT = 0;

}


void Worm::init( int num, int x_init, int y_init, int radius, int init_color, const unsigned char left, const unsigned char right, float spd, int ang )
{
	// color and size
	cbase = init_color;
	r = radius;

	// bearing
	speed = spd;
	angle = ang;
	dy = speed * sintab[angle];
	dx = speed * costab[angle];

	// inital location
	loc.x = x_init+dx;
	loc.y = y_init+dy;

	// queue up length
	Point p = { x_init, y_init };
	for ( int i=0; i<len; ++i )
		buf.enque(p);

	// make me unique
	ident = num;
	K_LEFT = left;
	K_RIGHT = right;

}


void Worm::draw()
{
	// draw my current location

	circlefill(screen,(int)loc.x,(int)loc.y,r,color);
	buf.enque(loc);
	old_loc = buf.deque();
	circlefill(screen,(int)old_loc.x,(int)old_loc.y,r,trail);

}


bool Worm::touches( Point p , int r2 )
{
	// check if given Point is within r2 of any of my component circles

	int max = buf.depth()-1;	// we skip our head
	Point current;
	Point diff;
	int i;

	float min_dist_sqr = (r+r2)*(r+r2);
	float hyp_sqr;

	for ( i=0; i<max; ++i )
	{
		current = buf[i];

		// get distance from current to p

		diff.x = current.x - p.x;
		diff.y = current.y - p.y;

		hyp_sqr = diff.x*diff.x + diff.y*diff.y;

		// now see if it's too close

		if ( hyp_sqr <= min_dist_sqr )
			return true;		// we have a winner

		// if not, we move on to the next circle in the queue
	}

	// if we're still here, it must not touch us.

	return false;
}


void Worm::grow(int diff)
{
	// get longer

	int max = buf.depth();

	// first enque my tail 'diff' times
	for ( int i=0; i<diff; ++i )
		buf.enque(buf[0]);
	// then, enque and deque my whole body to put new values at the end
	for ( int i=0; i<max; ++i )
		buf.enque( buf.deque() );

	len += diff;

}


bool Worm::update()
{
	// figure out where I should move to

	// rotate counter-clockwise
	if ( key[K_LEFT] )
	{
		angle = (angle+change)%360;
	}
	// rotate clockwise
	if ( key[K_RIGHT] )
	{
		if ( angle < change )
			angle += 360;
		angle -= change;
	}

	// and adjust dx and dy for the (possibly) new angle

	dy = speed * sintab[angle];
	dx = speed * costab[angle];

	// slow things down if there's friction

	if ( friction && speed > 0 )
		speed -= friction;

	// now update my location

	loc.x += dx;
	loc.y += dy;

	// check for collision with walls

	if ( loc.x-r < 0 )		// left wall
	{
		if ( WRAP )
			loc.x += WIDTH;
		else if ( BOUNCE )
			reflect(LEFT);
		else	// we die
			return true;
	}
	else if ( loc.x+r > WIDTH )	// right wall
	{
		if ( WRAP )
			loc.x -= WIDTH;
		else if ( BOUNCE )
			reflect(RIGHT);
		else	// we die
			return true;
	}
	else if ( loc.y-r < 0 )		// top wall
	{
		if ( WRAP )
			loc.y += HEIGHT;
		else if ( BOUNCE )
			reflect(TOP);
		else	// we die
			return true;
	}
	else if ( loc.y+r > HEIGHT )	// bottom wall
	{
		if ( WRAP )
			loc.y -= HEIGHT;
		else if ( BOUNCE )
			reflect(BOTTOM);
		else	// we die
			return true;
	}

	// and change my color

	color = cindex + cbase;
	cindex = (cindex+1)%6;

	return false;
}

void Worm::reflect(int face)
{
	switch(face)
	{
		case LEFT:
		case RIGHT: angle = 540 - angle;
		            loc.x -= dx;
		            if ( angle >= 360 ) angle -= 360;
		            break;
		case TOP:
		case BOTTOM: angle = 360 - angle;
		             loc.y -= dy;
		             break;
	}
	dy = speed * sintab[angle];
	dx = speed * costab[angle];
}


void Worm::relocate(int new_x, int new_y)
{
	loc.x = new_x;
	loc.y = new_y;
}


char* Worm::report()
{
	static char msg[50];
	sprintf(msg,"spd: %.1f ang: %d dx: %.4f dy: %.4f",speed,angle,dx,dy);
	return msg;
}

